"""Declares :class:`ServiceAccountPrincipal`."""
import dataclasses

from cbra.types import IPrincipal


@dataclasses.dataclass
class ServiceAccountPrincipal(IPrincipal):
    """Represents a principal that is a Google service account."""
    sub: str
    email: str

    def is_authenticated(self) -> bool:
        return True


"""Declares :func:`RFC9068Principal`."""
import inspect
import typing

import fastapi
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import ClaimSet
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials

from cbra.params import ServerKeychain
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import ForgedAccessToken
from cbra.exceptions import InvalidAuthorizationScheme
from cbra.exceptions import UntrustedIssuer
from .principal import Principal
from .types import IPrincipal
from .params import LocalIssuer


security = HTTPBearer(auto_error=False)


def RFC9068Principal(
    auto_error: bool = True,
    client_id: typing.Optional[str] = None,
    trusted_issuers: typing.Optional[typing.Set[str]] = None,
    header: str = 'Authorization',
    path: typing.Optional[str] = None,
    max_age: int = 300,
    principal_factory: typing.Callable[[ClaimSet], IPrincipal] = Principal.fromclaimset
) -> typing.Any:
    """Resolves the principal in a request to a subject using an
    :rfc:`9068` access token.
    """
    issuers: typing.Set[str] = trusted_issuers or set()
    if path is not None and not str.startswith(path, '/'):
        raise ValueError("The `path` parameter must start with a slash.")

    async def resolve_principal(
        request: fastapi.Request,
        issuer: str = LocalIssuer,
        bearer: typing.Optional[
            HTTPAuthorizationCredentials
        ] = fastapi.Depends(security),
        keychain: Keychain = ServerKeychain
    ):
        codec = PayloadCodec()
        if bearer is None:
            if auto_error and 'Authorization' not in request.headers:
                raise AuthenticationRequired
            if auto_error:
                # The header was present but a parsing error occurred.
                raise InvalidAuthorizationScheme
            return None

        if str.lower(bearer.scheme or '') != "bearer":
            raise InvalidAuthorizationScheme

        audience = {f"{request.url.scheme}://{request.url.netloc}{path or ''}"}
        jws, claims = await codec.jwt(bearer.credentials, accept="at+jwt")
        claims.verify(
            audience=audience,
            required={"jti", "iss", "aud", "sub", "iat", "nbf", "exp", "client_id"},
            max_age=max_age
        )
        if claims.iss not in (issuers | {issuer}):
            assert claims.iss is not None # nosec
            raise UntrustedIssuer(claims.iss)

        # If the issuer was valid but the signature validation fails here,
        # then the token was most probably forged.
        if claims.iss == issuer:
            is_valid_signature = jws.verify(keychain, require_kid=True)
        else:
            raise NotImplementedError
        if inspect.isawaitable(is_valid_signature):
            is_valid_signature = await is_valid_signature
        if not is_valid_signature:
            raise ForgedAccessToken
        return principal_factory(claims)

    return resolve_principal


# pylint: skip-file
import typing

import cbra
import cbra.exceptions
from cbra.ext.oauth2 import RFC9068Principal


class RFC9068Resource(cbra.Resource):
    name: str = 'resource'
    pluralname: str = 'resources'
    path_parameter: str = 'id:int'
    principal_factory = RFC9068Principal()

    async def authorize(self):
        if not self.principal.has_scope("read"):
            raise cbra.exceptions.NotAuthorized

    async def retrieve(self) -> typing.Dict[str, str]:
        return {
            "message": "Hello world!"
        }


"""Declares :class:`IEndpoint`."""
import typing

import fastapi
import pydantic
from aorta import MessagePublisher

from .irequesthandler import IRequestHandler


ResponseTypes = typing.Union[
    fastapi.Response,
    pydantic.BaseModel,
    typing.Dict[typing.Any, typing.Any],
    typing.List[typing.Any],
]


class IEndpoint(IRequestHandler):
    __module__: str = 'cbra'
    publisher: MessagePublisher
    ResponseTypes: typing.Any = ResponseTypes

    #: Automatically expose the ``OPTIONS`` method for this :class:`Endpoint`
    with_options: bool = False

    @classmethod
    def as_handler(
        cls,
        method: str,
        path: str,
        request_handler: typing.Optional[
            typing.Callable[..., ResponseTypes]
        ]
    ) -> typing.Coroutine[typing.Any, typing.Any, ResponseTypes]:
        raise NotImplementedError

    @classmethod
    def get_openapi_schema(cls) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def needs_body(cls, method: typing.Optional[str]) -> bool:
        raise NotImplementedError

    @classmethod
    def new(
        cls,
        **attrs: typing.Any
    ) -> typing.Type['IEndpoint']:
        """Create a new subclass of :class:`Endpoint` with the given
        attributes.
        """
        return type(cls.__name__, (cls,), attrs) # pragma: no cover

    @classmethod
    def update_signature(
        cls,
        runner: typing.Any,
        handle: typing.Any
    ) ->  typing.Coroutine[typing.Any, typing.Any, ResponseTypes]:
        raise NotImplementedError

    async def dispatch(
        self,
        handle: typing.Callable[..., ResponseTypes],
        **kwargs: typing.Any
    ) -> ResponseTypes:
        raise NotImplementedError

    async def handle(self) -> ResponseTypes:
        raise NotImplementedError

    async def render_to_response(
        self,
        content: typing.Any
    ) -> fastapi.Response:
        raise NotImplementedError


"""Declares :class:`ITokenIssuer`."""
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken

from .iclient import IClient
from .isubjectrepository import ISubjectRepository
from .servermetadata import ServerMetadata


class ITokenIssuer:
    __module__: str = 'cbra.ext.oauth2.types'

    #: Specifies the maximum age of an assertion.
    max_assertion_age: int = 30000

    #: Specifies the time-to-live of an access token if not otherwise
    #: specified, in seconds.
    default_ttl: int = 300

    #: The client that is handling the token request.
    client: IClient

    #: The concrete :class:`~cbra.ext.oauth2.ISubjectRepository`
    #: implementation that is used to lookup :term:`Subjects`.
    subjects: ISubjectRepository

    #: A :class:`~ckms.ext.oauth2.ServerMetadata` instance describing
    #: the authorization server that is issueing a token.
    metadata: ServerMetadata

    #: A :class:`ckms.jose.PayloadCodec` instance that knows how to
    #: decrypt JWE objects that were encrypted using the keys exposed
    #: at the JWKS URL.
    codec: PayloadCodec

    #: Indicates if a lookup of the JWKS of any issuer is allowed.
    allow_jwks_lookups: bool = True

    #: The set of trusted issuers for which the keys may be retrieved.
    trusted_issuers: set[str] = set()

    #: The URL of the local issuer i.e. the server. It is assumed that
    #: all tokens issued by this server have this value set to their
    #: issuer attribute.
    issuer: str

    def is_local_issued(self, claims: JSONWebToken) -> bool:
        raise NotImplementedError


"""Declares :class:`IStorage`."""
import functools
import typing

from ckms.types import ClaimSet

from ..exceptions import Error
from .authorizationcode import AuthorizationCode
from .authorizationrequest import AuthorizationRequest
from .baseauthorizationrequest import BaseAuthorizationRequest


class IStorage:
    """Declares the interface that storage implementations for the OAuth 2.0
    server need to implement.
    """
    __module__: str = 'cbra.ext.oauth2.types'
    InvalidAuthorizationCode: type[Error] = type('InvalidAuthorizationCode', (Error, ), {
        'error': "invalid_grant",
        'error_description': "The presented authorization code grant is not valid."
    })

    async def consume(self, claims: ClaimSet) -> bool:
        """Consume the claims set, ensuring that its token can only be
        used once. Return a boolean indicating if the token was already
        consumed.
        """
        raise NotImplementedError

    async def get(
        self,
        key: str
    ) -> typing.Optional[typing.Union[bytes, str]]:
        """Return a key from the transient storage backend, or ``None``
        if it does not exist.
        """
        raise NotImplementedError

    async def get_authorization_request(
        self,
        request_id: str
    ) -> AuthorizationRequest:
        """Lookup an authorization request from the transient storage."""
        raise NotImplementedError

    async def get_code(self, code: str) -> tuple[AuthorizationCode, AuthorizationRequest]:
        """Lookup an authorization code."""
        raise NotImplementedError

    async def get_pushed_authorization_request(
        self,
        request_uri: str
    ) -> BaseAuthorizationRequest:
        """Look an authorization request that was pushed through the Pushed
        Authorization Request (PAR) endpoint.
        """
        raise NotImplementedError

    async def set(
        self,
        key: str,
        value: typing.Union[bytes, str],
        expires: typing.Optional[int] = None
    ) -> None:
        """Set a `key` to given `value` to the **transient** storage backend.
        If `value` is a :class:`dict`, then all its members must be
        JSON-serializable.

        The `expires` parameter define the period after which the key expires
        i.e. it should be (automatically) removed from the transient storage
        backend. It is denoted in seconds.
        """
        raise NotImplementedError

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: AuthorizationRequest | AuthorizationCode
    ) -> None:
        """Persist an object to the transient storage."""
        raise NotImplementedError(type(obj).__name__)

    @persist.register
    async def _persist_authorizationcode(
        self,
        obj: AuthorizationCode
    ) -> None:
        return await self.persist_authorizationcode(obj)

    @persist.register
    async def _persist_authorizationrequest(
        self,
        obj: AuthorizationRequest
    ) -> None:
        return await self.persist_authorizationrequest(obj)

    async def persist_authorizationcode(
        self,
        obj: AuthorizationCode
    ) -> None:
        """Persist an :class:`cbra.ext.oauth2.types.AuthorizationCode`
        instance.
        """
        raise NotImplementedError

    async def persist_authorizationrequest(
        self,
        obj: AuthorizationRequest
    ) -> None:
        """Persist an :class:`cbra.ext.oauth2.AuthorizationRequest`
        instance.
        """
        raise NotImplementedError

    async def pop(
        self,
        key: str
    ) -> typing.Optional[typing.Union[bytes, str]]:
        """Like :meth:`get()`, but also removes the key."""
        raise NotImplementedError


"""Declares :class:`RequestContext`."""
import secrets
import typing

import fastapi


class RequestContext:
    __module__: str = 'cbra.ext.oauth2'

    @property
    def issuer(self) -> str:
        """The OAuth 2.0/OpenID Connect issuer, based on the
        current HTTP request.
        """
        return f'{self.request.url.scheme}://{self.request.url.netloc}'

    def __init__(self, request: fastapi.Request):
        self.request_id = secrets.token_bytes(24)
        self.request = request



# pylint: skip-file
import asyncio

import pytest
import pytest_asyncio
import ckms.core
from ckms.types import IKeySpecification


@pytest.fixture(scope='session')
def base_url() -> str:
    return "https://cbra.localhost"


@pytest.fixture(scope="session")
def event_loop():
    return asyncio.new_event_loop()


@pytest_asyncio.fixture(scope='session') # type: ignore
async def jose_client_signer() -> IKeySpecification:
    return await ckms.core.parse_spec({
        'provider': 'local',
        'kty': 'oct',
        'algorithm': 'HS256',
        'key': {'length': 32}
    })


"""Declates :class:`NamedQuery`."""
import inspect
import re
import typing

import pydantic

from cbra.ext.ioc import Dependency
from cbra.ext.ioc import Injected
from ..dependencywrapper import DependencyWrapper
from ..utils import PositionalArgument
from .namedqueryargs import NamedQueryArgs
from .namedquerymetaclass import NamedQueryMetaclass
from .queryrunner import QueryRunner


class NamedQuery(Dependency, metaclass=NamedQueryMetaclass):
    """A named query that is executed by a dynamically injected
    implementation of :class:`cbra.QueryRunner`.

    The `runner` arguments point to a dependency that is injected at runtime
    under the given key. It is expected to implement :class:`cbra.QueryRunner`
    interface. The default value is ``'QueryRunner'``.
    """
    __abstract__: bool = True
    __module__: str = 'cbra.query'
    default_runner: str = 'default'

    @classmethod
    def get_method_name(cls) -> str:
        """Convert class name `name` to underscored method name."""
        return str.lower(re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__))

    def __init__(self, using: typing.Optional[str] = None, run: bool = True):
        self.must_run = run
        self.using = using or self.default_runner
        self.resolve = DependencyWrapper(self._run, [
            PositionalArgument(
                name='runner',
                default=Injected(
                    f'cbra.runner.{self.using}',
                    invoke=True,
                    default=None
                ),
                annotation=QueryRunner
            )
        ] + list(inspect.signature(self.build).parameters.values()))
        super().__init__(use_cache=True)

    async def execute(self, runner: QueryRunner, query: NamedQueryArgs):
        """Invokes the appropriate method on the :class:`cbra.QueryRunner`
        `runner` instance.
        """
        return await getattr(runner, f'run_{self.get_method_name()}')(query)

    def build(self) -> pydantic.BaseModel:
        """Build the query model from the given parameters."""
        raise NotImplementedError

    async def _run(self, runner: QueryRunner, *args, **kwargs):
        query = self.build(*args, **kwargs)
        if not self.must_run:
            return query

        return await self.execute(runner, query)



"""Declares :class:`AnonymousReadCorsPolicy`."""
from .basecorspolicy import BaseCorsPolicy


class AnonymousReadCorsPolicy(BaseCorsPolicy):
    __module__: str = 'cbra.cors'
    allowed_methods: set[str] = {"GET"}
    allowed_headers: set[str] = {"Wants-Digest"}
    allowed_response_headers: set[str] = {"Digest"}

    async def get_allowed_origins(self) -> set[str]:
        if not self.origin:
            return set()
        return {self.origin}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationrequest import AuthorizationRequest
from .basegrant import BaseGrant
from .iclient import IClient
from .isubject import ISubject


class IOpenIdTokenBuilder:
    __module__: str = 'cbra.ext.oauth2.types'

    async def build(
        self,
        *,
        signing_key: str,
        client: IClient,
        subject: ISubject,
        grant: BaseGrant,
        scope: set[str],
        request: AuthorizationRequest | None = None,
        access_token: str | None = None
    ) -> str | None:
        """Build an OpenID Connect ID Token."""
        raise NotImplementedError


# pylint: skip-file
import itertools
import json
import typing
import urllib.parse

import httpx
import pytest
import pytest_asyncio
import yaml

import cbra
from examples.resource import get_asgi_application
from examples.resource import BarExampleResourceModel
from examples.resource import ExampleResource
from examples.resource import FooExampleResourceModel


VALID_REQUESTS = [
    BarExampleResourceModel(bar=1),
    FooExampleResourceModel(foo=2)
]

SUPPORTED_CONTENT_TYPES = [
    ("application/json", json.dumps),
    ("application/yaml", yaml.safe_dump),
    ("application/x-www-form-urlencoded", urllib.parse.urlencode),
]


class ResourceTestCase:
    allowed_detail_metdhods: typing.Set[str] = {
        "GET",
        "PATCH",
        "PUT",
        "DELETE",
        "OPTIONS"
    }
    allowed_list_methods: typing.Set[str] = {
        "POST",
        "GET",
        "DELETE",
        "OPTIONS"
    }
    detail_url: str = '/resources/1'
    list_url: str = '/resources'
    path_params = {
        'id': 1
    }

    @pytest.fixture
    def app(self) -> cbra.Application:
        return get_asgi_application()

    @pytest_asyncio.fixture # type: ignore
    async def client(
        self,
        app: cbra.Application
    ) -> typing.AsyncGenerator[httpx.AsyncClient, None]:
        params = {
            'base_url': "https://cbra.localhost",
            'headers': {
                'Accept': "application/json"
            }
        }
        async with httpx.AsyncClient(app=app, **params) as c:
            yield c

    @pytest.mark.parametrize("method", allowed_detail_metdhods)
    @pytest.mark.asyncio
    async def test_resource_handle_coerces_parameters(
        self,
        client: httpx.AsyncClient,
        method: str
    ):
        response = await client.get(self.detail_url)
        assert response.status_code == 200, response.text
        dto = response.json()
        assert 'id' in dto
        assert isinstance(dto['id'], int)

    @pytest.mark.parametrize("method", allowed_detail_metdhods)
    @pytest.mark.asyncio
    async def test_resource_handle_sees_parameters(
        self,
        client: httpx.AsyncClient,
        method: str
    ):
        response = await client.get(self.detail_url)
        assert response.status_code == 200
        dto = response.json()
        assert 'path_params' in dto
        assert set(dto['path_params'].keys()) == set(self.path_params.keys())


    @pytest.mark.parametrize("method", allowed_list_methods)
    @pytest.mark.asyncio
    async def test_list_handle_sees_parameters(
        self,
        client: httpx.AsyncClient,
        method: str
    ):
        response = await client.get(self.list_url)
        assert response.status_code == 200, response.text
        dto = response.json()
        assert 'path_params' in dto
        seen = set(dto['path_params'].keys())
        available = set(self.path_params.keys())
        assert seen < available, dto
        assert len(seen) == len(available) - 1

    @pytest.mark.asyncio
    async def test_options_list_exists(
        self,
        client: httpx.AsyncClient,
    ):
        response = await client.options(self.list_url)
        assert response.status_code == 200, response.status_code
        assert 'Allow' in response.headers, response.headers
        allowed = set([x.strip() for x in response.headers['Allow'].split(',')])
        assert allowed == self.allowed_list_methods, allowed

    @pytest.mark.asyncio
    async def test_options_detail_exists(
        self,
        client: httpx.AsyncClient,
    ):
        response = await client.options(self.detail_url)
        assert response.status_code == 200
        allowed = set([x.strip() for x in response.headers['Allow'].split(',')])
        assert allowed == self.allowed_detail_metdhods, allowed

    @pytest.mark.parametrize("dto", [
        BarExampleResourceModel(bar=1),
        FooExampleResourceModel(foo=2),
    ])
    @pytest.mark.asyncio
    async def test_create_multiple_example(
        self,
        client: httpx.AsyncClient,
        dto: typing.Union[BarExampleResourceModel, FooExampleResourceModel]
    ):
        response = await client.post(
            url=self.list_url,
            json=dto.dict()
        )
        result = response.json()
        assert response.status_code == 201
        assert 'dto' in result
        assert dto.dict() == result['dto']

    @pytest.mark.parametrize("dto", [
        BarExampleResourceModel(bar=1),
        FooExampleResourceModel(foo=2),
    ])
    @pytest.mark.asyncio
    async def test_replace_multiple_example(
        self,
        client: httpx.AsyncClient,
        dto: typing.Union[BarExampleResourceModel, FooExampleResourceModel]
    ):
        response = await client.put(
            url=self.detail_url,
            json=dto.dict()
        )
        result = response.json()
        assert response.status_code == 200
        assert 'dto' in result
        assert dto.dict() == result['dto']

    @pytest.mark.parametrize("dto", [
        {'123': 'abc'},
        []
    ])
    @pytest.mark.asyncio
    async def test_create_rejects_invalid_body(
        self,
        client: httpx.AsyncClient,
        dto: typing.Any
    ):
        response = await client.post(
            url=self.list_url,
            json=dto
        )
        assert response.status_code == 422

    @pytest.mark.parametrize("dto", [
        {'123': 'abc'},
        []
    ])
    @pytest.mark.asyncio
    async def test_replace_rejects_invalid_body(
        self,
        client: httpx.AsyncClient,
        dto: typing.Any
    ):
        response = await client.put(
            url=self.detail_url,
            json=dto
        )
        assert response.status_code == 422

    @pytest.mark.asyncio
    async def test_query_parses_parameters(
        self,
        client: httpx.AsyncClient
    ):
        pytest.skip()
        params = ExampleResource.query(foo=1, bar=2, baz=['3', '4'])
        response = await client.get(
            url=self.list_url,
            params=params.dict()
        )
        assert response.status_code == 200


"""Declares :class:`MemoryStorage`."""
import typing

from ckms.types import ClaimSet

from .types import AuthorizationCode
from .types import AuthorizationRequest
from .types import IStorage


class MemoryStorage(IStorage):
    """A :class:`~cbra.ext.oauth2.IStorage` implementation
    that used the local memory. For testing purposes only.
    """
    __module__: str = 'cbra.ext.oauth2'
    objects: typing.Dict[str, typing.Any | AuthorizationCode | AuthorizationRequest] = {}

    def __init__(self):
        self.objects = MemoryStorage.objects

    @classmethod
    def clear(cls) -> None:
        MemoryStorage.objects = {}

    async def consume(self, claims: ClaimSet) -> bool:
        """Consume the claims set, ensuring that its token can only be
        used once. Return a boolean indicating if the token was already
        consumed.
        """
        k = f'consumed.{claims.jti}'
        is_consumed = True
        if k not in self.objects:
            self.objects[k] = ...
            is_consumed = False
        return is_consumed

    async def get_authorization_request(self, request_id: str) -> AuthorizationRequest:
        return self.objects[request_id]

    async def get_code(self, code: str) -> tuple[AuthorizationCode, AuthorizationRequest]:
        try:
            obj = typing.cast(AuthorizationCode, self.objects[code])
            return obj, self.objects[obj.request_id]
        except KeyError:
            raise self.InvalidAuthorizationCode

    async def persist_authorizationrequest(
        self,
        obj: AuthorizationRequest
    ) -> None:
        assert obj.request_id # nosec
        self.objects[obj.request_id] = obj

    async def persist_authorizationcode(self, obj: AuthorizationCode) -> None:
        assert obj.code
        self.objects[obj.code] = obj


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`OrderEvent`."""
import pydantic

from .event import Event
from .orderstatustype import OrderStatusType


class OrderEvent(Event):
    status: OrderStatusType = pydantic.Field(
        default=...,
        title="Status",
        description="The status of the order resulting from the event."
    )

    class Config:
        title: str = "WooCommerceOrderEvent"


"""Declares :class:`MetadataEndpoint`."""
import cbra

from .types import IAuthorizationServer
from .types import ServerMetadata


class MetadataEndpoint(cbra.Endpoint):
    __module__: str = 'cbra.ext.oauth2'
    require_authentication: bool = False
    summary: str = "Server metadata"
    description: str = (
        "The metadata endpoint returns a JSON datastructure that an OAuth 2.0 "
        "client can use to obtain the information needed to interact with an "
        "OAuth 2.0 authorization server, including its endpoint locations and "
        "authorization server capabilities."
    )
    method: str = "GET"
    response_description: str = (
        "A JSON datastructure describing the authorization server "
        "(see RFC 8414)."
    )
    response_model: type[ServerMetadata] = ServerMetadata
    server: IAuthorizationServer
    with_options: bool = True

    async def handle(self) -> ServerMetadata:
        return await self.server.get_metadata(self.request)


# pylint: skip-file
import urllib.parse

import httpx

from .iauthorizationcodeflow import IAuthorizationCodeFlow
from .test_unit_jar import TestJARFlow as Base


class TestRequestParameter(Base):

    async def create_authorize_response(
        self,
        client: httpx.AsyncClient,
        params: dict[str, str],
        status_code: int = 303,
        error: str | None = None,
        headers: dict[str, str] | None = None
    ) -> tuple[urllib.parse.ParseResult, dict[str, str], dict[str, str]]:
        params = {
            'client_id': self.client_id,
            'request': (
                await self.create_jar(
                    codec=await self.get_client_codec(
                        keychain=await self.get_client_keychain(self.client_key)
                    ),
                    params=params
                )
            )
        }
        return await IAuthorizationCodeFlow.create_authorize_response(
            self,
            client=client,
            params=params,
            status_code=status_code,
            error=error
        )


class TestRequestParameterCORS(TestRequestParameter):
    origin = "https://allowed-cors-domain.example"


# pylint: skip-file
import random
import typing
import urllib.parse

import httpx
import pytest
from ckms.jose import PayloadCodec
from unimatrix.lib.http import parse_qs # type: ignore

import cbra
from cbra.ext.oauth2 import AuthorizationServer
from ...corsmixin import CorsMixin


class IAuthorizationCodeFlow(CorsMixin):
    authorize_url: str
    base_url: str
    client_id: str
    client_key: str
    cors_allowed_origin: str = "https://allowed-cors-domain.example"
    cors_allowed_methods: set[str] = {"GET"}
    origin: str | None = None
    grant_type: str = "authorization_code"
    redirect_uri: str
    redirects: bool = False
    token_url: str
    authorize_status_code: int
    response_type: str

    async def get_token(
        self,
        client: httpx.AsyncClient,
        code: str
    ) -> dict[str, typing.Any]:
        headers = {}
        if self.origin:
            headers['Origin'] = self.origin
            preflight = await client.options( # type: ignore
                url=self.token_url,
                headers={
                    'Origin': self.origin,
                    'Access-Control-Request-Methods': "POST"
                }
            )
            assert preflight.status_code == 200, preflight.status_code
            assert preflight.headers.get('Access-Control-Allow-Origin') == self.origin, preflight.headers
            assert preflight.headers.get('Access-Control-Allow-Credentials') == "false"
            assert "POST" in preflight.headers.get('Access-Control-Allow-Methods', ''), preflight.headers

        response = await client.post( # type: ignore
            url=self.token_url,
            json=await self.get_token_request_params(code),
            headers=headers
        )
        assert response.status_code == 200, response.text
        if self.origin is not None:
            assert 'Access-Control-Allow-Origin' in response.headers, response.headers
            assert response.headers['Access-Control-Allow-Origin'] == self.origin
        return response.json()

    async def get_token_request_params(self, code: str) -> dict[str, typing.Any]:
        return {
            'grant_type': self.grant_type,
            'code': code,
            'redirect_uri': self.redirect_uri,
            **await self.get_token_client_params()
        }

    async def get_token_client_params(self) -> dict[str, typing.Any]:
        return {'client_id': self.client_id}

    async def create_authorize_response(
        self,
        client: httpx.AsyncClient,
        params: dict[str, str],
        status_code: int = 303,
        error: str | None = None,
        headers: dict[str, str] | None = None
    ) -> tuple[urllib.parse.ParseResult, dict[str, str], dict[str, str]]:
        headers = headers or {}
        if self.origin is not None:
            headers['Origin'] = self.origin
        response = await client.get( # type: ignore
            url=self.authorize_url,
            params=params,
            headers=headers,
        )
        if 'Location' not in response.headers:
            dto = response.json()
            assert response.status_code == status_code, f'{response.status_code}: {dto}'
            assert response.headers['Content-Type'] == "application/json"
            assert 'error' in dto
            if error is not None:
                assert dto['error'] == error, dto
            return (urllib.parse.urlparse(''), dict(response.headers), {})
        
        assert response.status_code == status_code, f'{response.status_code}: {response.text}'
        if self.origin is not None:
            assert 'Access-Control-Allow-Origin' in response.headers
            assert response.headers['Access-Control-Allow-Origin'] == self.origin
        p = urllib.parse.urlparse(response.headers['Location'])
        dto = dict(parse_qs(p.query))
        if error is not None:
            assert dto.get('error') == error, dto
        return (
            p,
            dict(response.headers),
            dto
        )

    async def get_minimal_authorize_params(self, **params: typing.Any) -> dict[str, str]:
        return {
            'client_id': self.client_id,
            'response_type': 'code',
            'redirect_uri': self.redirect_uri,
            **params
        }

    @pytest.mark.asyncio
    async def test_authorize_with_minimal_parameters(
        self,
        client: httpx.AsyncClient,
        headers: dict[str, str] | None = None,
    ) -> dict[str, str]:
        url, headers, result = await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(),
            status_code=303,
            headers=headers
        )
        assert url[:3] == urllib.parse.urlparse(self.redirect_uri)[:3]
        assert 'code' in result, result
        assert 'iss' in result, result
        assert result['iss'] == self.base_url
        token = await self.get_token(
            client=client,
            code=typing.cast(str, result.get('code'))
        )
        assert 'access_token' in token
        assert 'token_type' in token
        assert token['token_type'] == 'Bearer'

        header, claims = PayloadCodec().introspect(token['access_token'])
        assert claims is not None
        assert claims.extra['client_id'] == self.client_id
        assert claims.sub == client.cookies['subject']
        assert header.headers[0].typ == "at+jwt"
        return headers

    @pytest.mark.asyncio
    async def test_authorize_with_minimal_parameters_cors(
        self,
        client: httpx.AsyncClient,
        headers: dict[str, str] | None = None
    ):
        origin = "https://allowed-cors-domain.example"
        await self.test_authorize_with_minimal_parameters(
            client=client,
            headers={
                'Origin': origin
            }
        )

    @pytest.mark.asyncio
    async def test_authorize_with_minimal_parameters_retains_state(
        self,
        client: httpx.AsyncClient
    ) -> None:
        *_, result = await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                state='foo'
            ),
            status_code=303
        )
        assert 'state' in result
        assert result.get('state') == 'foo'

    @pytest.mark.asyncio
    async def test_authorize_redirect_error_retains_state(
        self,
        client: httpx.AsyncClient
    ) -> None:
        # Trigger some error that can redirect to the client. Do not run
        # this test on pushed authorization request test cases.
        if not self.redirects or self.origin is not None:
            pytest.skip()
        *_, result = await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                state='foo'
            ),
            headers={'Origin': 'https://foo.invalid'},
            error='invalid_origin',
            status_code=303,
        )
        assert 'state' in result
        assert result.get('state') == 'foo'

    @pytest.mark.asyncio
    async def test_authorize_retains_query(
        self,
        client: httpx.AsyncClient
    ) -> None:
        *_, result = await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri="https://cbra-oauth2.localhost/callback?foo=bar"
            ),
            status_code=303
        )
        assert 'foo' in result
        assert result.get('foo') == 'bar'

    @pytest.mark.asyncio
    async def test_authorize_unregistered_query_parameter(
        self,
        client: httpx.AsyncClient
    ) -> None:
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri="https://cbra-oauth2.localhost/callback?bar=foo"
            ),
            status_code=400,
            error="invalid_request"
        )

    @pytest.mark.asyncio
    async def test_authorize_unregistered_query_value(
        self,
        client: httpx.AsyncClient
    ) -> None:
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri="https://cbra-oauth2.localhost/callback?foo=baz"
            ),
            status_code=400,
            error="invalid_request"
        )

    @pytest.mark.asyncio
    async def test_authorize_rejects_url_with_fragment(
        self,
        client: httpx.AsyncClient
    ) -> None:
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri=f'{self.redirect_uri}#foo'
            ),
            status_code=400,
            error="invalid_request"
        )

    @pytest.mark.asyncio
    async def test_authorize_rejects_url_not_registered(
        self,
        client: httpx.AsyncClient
    ) -> None:
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri=f'https://www.example.com'
            ),
            status_code=400,
            error="invalid_request"
        )

    @pytest.mark.asyncio
    async def test_authorize_rejects_non_https(
        self,
        client: httpx.AsyncClient
    ) -> None:
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri=f'http://cbra-oauth2.localhost/callback'
            ),
            status_code=400,
            error="invalid_request"
        )

    @pytest.mark.asyncio
    async def test_authorize_allows_http_on_loopback(
        self,
        client: httpx.AsyncClient
    ) -> None:
        # Authorization servers MUST require clients to register their complete
        # redirect URI (including the path component) and reject authorization
        # requests that specify a redirect URI that doesn't exactly match the
        # one that was registered; the exception is loopback redirects, where
        # an exact match is required except for the port URI component.
        # (OAuth 2.1 draft, 9.2)
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri=f"http://127.0.0.1:{random.randint(1000,10000)}"
            ),
            status_code=303
        )

        # The path /callback is not whitelisted in the test client, so
        # this request must fail.
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                redirect_uri=f"http://127.0.0.1:{random.randint(1000,10000)}/callback"
            ),
            status_code=400,
            error="invalid_request"
        )


    @pytest.mark.asyncio
    async def test_openid_authorize_fails_with_login_required_when_prompt_is_none(
        self,
        client: httpx.AsyncClient
    ) -> None:
        client.cookies.pop('subject', None) # type: ignore
        await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(
                scope="openid",
                prompt="none"
            ),
            status_code=303,
            error="login_required"
        )

    @pytest.mark.asyncio
    async def test_authorize_unauthenticated_redirects_to_login_url(
        self,
        server: AuthorizationServer,
        client: httpx.AsyncClient
    ):
        client.cookies.pop('subject', None) # type: ignore
        server.login_url = "https://www.example.com/login"
        _, headers, query = await self.create_authorize_response(
            client=client,
            params=await self.get_minimal_authorize_params(),
            status_code=303,
        )
        assert 'location' in headers
        assert headers['location'].startswith(server.login_url)
        assert 'next' in query
        assert query['next'].startswith(self.authorize_url)


"""Declares :class:`NullPrincipal`."""
from cbra.types import IPrincipal
from cbra.utils import docstring


class NullPrincipal(IPrincipal):
    """An :class:`~cbra.types.IPrincipal` implementation that is used
    with non-authenticated requests.

    This is the default principal for all endpoints that did not specify
    otherwise.
    """
    __module__: str = 'cbra.auth'

    def __init__(self):
        self.sub = None

    @docstring(IPrincipal)
    def is_authenticated(self) -> bool:
        return False

    def __bool__(self) -> bool:
        return False


"""Declares :class:`BrowserRenderer`."""
import typing

from .yaml import YAMLRenderer


class BrowserRenderer(YAMLRenderer):
    """Like :class:`YAMLRenderer`, but uses ``text/plain`` as output
    since most browsers don't support YAML.
    """
    media_type: str = 'text/html'
    exact: bool = False

    def render(self, *args, **kwargs) -> typing.Any: # type: ignore
        content = super().render(*args, **kwargs)
        return f'<html><head></head><body><pre>{content}</pre></body></html>'


# pylint: skip-file
import typing


__all__ = [
    'REQUEST_METHODS'
]


REQUEST_METHODS: typing.Set[str] = {
    "GET",
    "POST",
    "PUT",
    "PATCH",
    "OPTIONS",
    "DELETE",
    "TRACE",
    "HEAD"
}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

from .types import IClient
from .types import ISubject


class OIDCClaimHandler:
    """Provides an interface to update a claims set for a given
    scope.
    """
    __module__: str = 'cbra.ext.oauth2'
    client: IClient
    handles: set[str] = set()
    now: int
    subject: ISubject

    def __init__(
        self,
        client: IClient,
        subject: ISubject,
        now: int,
        initial: dict[str, Any]
    ):
        self.client = client
        self.now = now
        self.subject = subject

    async def claims(self) -> dict[str, Any]:
        """Return a mapping holding the claims that are added to the
        claims set by this handler.
        """
        raise NotImplementedError

    @functools.singledispatchmethod
    def is_enabled(self, config: set[str]) -> bool:
        """Return a boolean if the handler is enabled for the given
        scope or claims `config.
        """
        raise NotImplementedError(type(config))

    @is_enabled.register
    def _is_enabled_for_scope(self, scope: set) -> bool: # type: ignore
        return self.is_enabled_for_scope(scope) # type: ignore

    def is_enabled_for_scope(self, scope: set[str]) -> bool:
        """Return a boolean if the handler is enabled for the given
        scope `scope`.
        """
        return set(scope) >= set(self.handles)


"""Declares :class:`Path`."""
import pydantic.fields


class PathField(pydantic.fields.FieldInfo):
    """A :class:`pydantic.Field` implementation that indicates that
    a parameter comes from the request path component.
    """
    pass



"""Declares :class:`NullRenderer`."""
import warnings
from typing import Any

import pydantic

from cbra.types import IRenderer


class NullRenderer(IRenderer):
    """
    Renderer which serializes to JSON.
    """
    __module__: str = 'cbra.renderers'

    def has_content(self) -> bool:
        return False

    def render(
        self,
        data: dict[str, Any] | pydantic.BaseModel,
        renderer_context: dict[str, Any] | None = None
    ) -> bytes:
        """Return a an empty byte-string, indicating no response body."""
        if bool(data):
            warnings.warn(f"{type(self).__name__} can not render any content.")
        return b''


"""Declares :class:`AuthorizationRequest`."""
from .baseauthorizationrequest import BaseAuthorizationRequest


Base = BaseAuthorizationRequest.query()


class AuthorizationRequestParameters(Base):

    @property
    def request_id(self) -> str | None:
        """Return the request identified as specified by the :attr:`request_uri`
        parameter.
        """
        if self.request_uri is None\
        or not str.startswith(self.request_uri, 'urn:ietf:params:oauth:request_uri'):
            return None
        *_, request_id = str.rsplit(self.request_uri, ':', 1)
        return request_id

    def is_external(self) -> bool:
        """If the request is a reference, return a boolean indicating if the
        request is made available at an external source.
        """
        return not str.startswith(
            self.request_uri or '',
            "urn:ietf:params:oauth:request_uri"
        )

    def is_object(self) -> bool:
        """Return a boolean indicating if the parameters contain a request
        object, supplied by the :attr:`request` parameter.
        """
        return self.request is not None

    def is_reference(self) -> bool:
        """Return a boolean indicating if the authorization request is a
        reference.
        """
        return self.request_uri is not None

    async def resolve(self):
        """Resolves the parameters based on the ``request`` and
        ``request_uri`` parameters.
        """
        pass


"""Declares :class:`fastapi.params.Header` instances for
common HTTP headers.
"""
import typing

import fastapi
import fastapi.params


ACCEPT: str = fastapi.Header(
    default=None,
    title="Accept",
    alias="Accept",
    description=(
        "Indicates which content types, expressed as MIME "
        "types, the client is able to understand."
    )
)

CONTENT_ENCODING_SCHEMA: dict[str, typing.Any] = {
    'description': (
        "Lists any encodings that have been applied to the representation "
        "(message payload), and in what order. This lets the recipient "
        "know how to decode the representation in order to obtain the "
        "original payload format"
    ),
    'schema': {
        'type': 'string',
        'example': 'deflate, gzip'
    }
}

CONTENT_ENCODING: str = fastapi.Header(
    default=None,
    alias="Content-Encoding",
    title="Content-Encoding",
    description=CONTENT_ENCODING_SCHEMA['description']
)

CONTENT_LENGTH_SCHEMA: dict[str, typing.Any] = {
    'description': (
        "Indicates the size of the message body, in bytes, sent to the recipient."
    ),
    'schema': {
        'type': 'number',
        'example': 1
    }
}

CONTENT_LENGTH: int = fastapi.Header(
    default=None,
    alias="Content-Length",
    title="Content-Length",
    description=CONTENT_LENGTH_SCHEMA['description']
)

CONTENT_LOCATION_SCHEMA: dict[str, typing.Any] = {
    'description': (
        "Indicates an alternate location for the returned data. "
        "The principal use is to indicate the URL of a resource "
        "transmitted as the result of content negotiation. It "
        "may also be used to indicate the location of a resource "
        "resulting from its creation."
    ),
    'schema': {
        'type': 'string',
        'example': "https://api.example.com/1"
    }
}

CONTENT_TYPE_SCHEMA: dict[str, typing.Any] = {
    'description': (
        "Used to indicate the original media type of the resource "
        "(prior to any content encoding applied for sending)."
    ),
    'schema': {
        'type': 'string',
        'example': 'application/json'
    }
}

CONTENT_TYPE: str = fastapi.Header(
    default=None,
    alias="Content-Type",
    title="Content-Type",
    description=CONTENT_TYPE_SCHEMA['description']
)


ACCEPT_LANGUAGE_SCHEMA: dict[str, typing.Any] = {
    'description': (
        "Indicates the natural language and locale that the client prefers. "
        "The server uses content negotiation to select one of the proposals "
        "and informs the client of the choice with the `Content-Language` "
        "response header."
    ),
    'schema': {
        'type': 'string',
        'example': 'fr-CH, fr;q=0.9, en;q=0.8, de;q=0.7, *;q=0.5'
    }
}

ACCEPT_LANGUAGE: str = fastapi.Header(
    default=None,
    title="Accept-Language",
    alias="Accept-Language",
    description=ACCEPT_LANGUAGE_SCHEMA['description']
)


ACCESS_CONTROL_ALLOW_CREDENTIALS_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Tells browsers whether to expose the response to the "
        "frontend JavaScript code when the request's credentials "
        "mode (`Request.credentials`) is `include`.\n\n"
        "Credentials are cookies, authorization headers, "
        "or TLS client certificates."
    ),
    'schema': {
        'type': 'boolean',
        'example': 'false'
    }
}

ACCESS_CONTROL_ALLOW_HEADERS_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Used in response to a preflight request which includes the "
        "`Access-Control-Request-Headers` to indicate which HTTP "
        "headers can be used during the actual request."
    ),
    'schema': {
        'type': 'string',
        'example': 'X-Custom-Header, X-Another-Custom-Header'
    }
}

ACCESS_CONTROL_ALLOW_METHODS_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Specifies one or more methods allowed when accessing a "
        "resource in response to a preflight request."
    ),
    'schema': {
        'type': 'string',
        'example': 'GET, OPTIONS'
    }
}

ACCESS_CONTROL_ALLOW_ORIGIN_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Indicates whether the response can be shared "
        "with requesting code from the origin specified "
        "in the `Origin` request header."
    ),
    'schema': {
        'type': 'string',
        'example': '*'
    }
}

ACCESS_CONTROL_EXPOSE_HEADERS_SCHEMA: typing.Dict[str, typing.Any] = {
    'description': (
        "Indicate which response headers should be made available to "
        "scripts running in the browser, in response to a cross-origin "
        "request.\n\n"
        "Only the CORS-safelisted response headers are exposed by "
        "default. Additional headers are listed by this header."
    ),
    'schema': {
        'type': 'string',
        'example': 'Content-Encoding, X-Custom-Header'
    }
}

ACCESS_CONTROL_MAX_AGE_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Indicates how long the results of a preflight request "
        "(that is the information contained in the "
        "`Access-Control-Allow-Methods` and "
        "`Access-Control-Allow-Headers` headers) can be cached."
    ),
    'schema': {
        'type': 'number',
        'example': '600'
    }
}

ACCESS_CONTROL_REQUEST_HEADERS: str = fastapi.Header(
    default=None,
    alias='Access-Control-Request-Headers',
    title="Access-Control-Request-Headers",
    description=(
        "Used by browsers when issuing a preflight request to let "
        "the server know which HTTP headers the client might send "
        "when the actual request is made. The complementary "
        "server-side header of `Access-Control-Allow-Headers` will "
        "answer this browser-side header."
    )
)

ACCESS_CONTROL_REQUEST_METHOD: str = fastapi.Header(
    default=None,
    alias='Access-Control-Request-Methods',
    title="Access-Control-Request-Methods",
    description=(
        "Used by browsers when issuing a preflight request, to let "
        "the server know which HTTP method will be used when the "
        "actual request is made. This header is necessary as the "
        "preflight request is always an `OPTIONS` and doesn't use "
        "the same method as the actual request."
    )
)

CORS_HEADERS: typing.Dict[str, typing.Any] = {
    'Access-Control-Allow-Credentials': ACCESS_CONTROL_ALLOW_CREDENTIALS_SCHEMA,
    'Access-Control-Allow-Origin': ACCESS_CONTROL_ALLOW_ORIGIN_SCHEMA,
    'Access-Control-Expose-Headers': ACCESS_CONTROL_EXPOSE_HEADERS_SCHEMA
}

DIGEST_SCHEMA: typing.Dict[
    str,
    typing.Union[str, typing.Dict[str, str]]
] = {
    'description': (
        "Provides a digest of the *selected representation* of the requested "
        "resource.\n\nRepresentations are different forms of a particular "
        "resource that might be returned from a request: for example, the "
        "same resource might be formatted in a particular media type such "
        "as XML or JSON, localized to a particular written language or "
        "geographical region, and/or compressed or otherwise encoded for "
        "transmission. The *selected representation* is the actual format of "
        "a resource that is returned following content negotiation, and can "
        "be determined from the response's Representation headers.\n\n"
        "The digest applies to the whole representation of a resource, not "
        "to a particular message. It can be used to verify that the "
        "representation data has not been modified during transmission."
    ),
    'schema': {
        'type': 'string'
    }
}

ORIGIN: str | None = fastapi.Header(
    default=None,
    alias='Origin',
    title="Origin",
    description=(
        "Indicates the origin (scheme, hostname, and port) that "
        "caused the request. For example, if a user agent needs "
        "to request resources included in a page, or fetched by "
        "scripts that it executes, then the origin of the page "
        "may be included in the request."
    )
)


WANTS_DIGEST: str = fastapi.Header(
    default=None,
    title="Wants digest",
    alias="Wants-Digest",
    description=(
        "Instructs the the server to provide a digest of the "
        "requested resource using the `Digest` response header."
    ),
    example="SHA-512;q=0.3, sha-256;q=1"
)


HEADERS_RESPONSE_BODY: dict[str, typing.Any] = {
    'Content-Encoding': CONTENT_ENCODING_SCHEMA,
    'Content-Length': CONTENT_LENGTH_SCHEMA,
    'Content-Location': CONTENT_LOCATION_SCHEMA,
    'Content-Type': CONTENT_TYPE_SCHEMA,
}


# pylint: skip-file
import datetime
import functools
import inspect
import itertools
import typing
from inspect import Parameter

from .mediatypes import media_type_matches
from .mediatypes import parse_header


__all__ = [
	'media_type_matches',
    'normalize_signature',
    'parse_header',
]


class DeferredException:
    """Defer raising the exception until an attribute is accessed."""
    exc: BaseException

    def __init__(self, exc: BaseException):
        self.exc = exc

    def raise_exception(self) -> typing.NoReturn:
        raise self.exc

    def __getattr__(self, attname: str) -> typing.NoReturn:
        raise self.exc

    __bytes__ = raise_exception
    __str__ = raise_exception


def current_timestamp() -> int:
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())


def docstring(cls: type[typing.Any]): # type: ignore
    def update_func(func):
        func.__doc__ = getattr(cls, func.__name__).__doc__
        return func
    return update_func


def order_parameters(params: typing.List[Parameter]) -> typing.List[Parameter]:
    """Take a list of parameters and ensure that they are ordered correctly i.e.
    positional arguments, keyword arguments, variable positional arguments and
    variable keyword arguments.
    """
    default = [x for x in params if x.default != Parameter.empty]
    nondefault = [x for x in params if x.default == Parameter.empty]
    result = itertools.chain(
        nondefault,
        [x for x in default if x .kind == Parameter.POSITIONAL_ONLY],
        [x for x in default if x .kind == Parameter.POSITIONAL_OR_KEYWORD],
        [x for x in default if x .kind == Parameter.VAR_POSITIONAL],
        [x for x in default if x .kind == Parameter.KEYWORD_ONLY],
        [x for x in default if x .kind == Parameter.VAR_KEYWORD],
    )
    return list(result)


def normalize_signature(func):
    """Normalize the signature of a function or method so that it
    can be used as a :mod:`fastapi` request handler.
    """
    sig = inspect.signature(func)
    func.__signature__ = sig.replace(
        parameters=[
            x for x in sig.parameters.values()
            if x.name != 'self'
        ]
    )
    return func


class PositionalArgument(inspect.Parameter):

    def __init__(self,
        name: str,
        annotation: type = None,
        **kwargs
    ):
        super().__init__(
            name=name,
            kind=Parameter.POSITIONAL_ONLY,
            annotation=annotation,
            **kwargs
        )


class ClassPropertyDescriptor:

    @classmethod
    def new(cls, func: typing.Callable[..., typing.Any]) -> typing.Any:
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        return cls(func)

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None) -> typing.Any:
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value) -> typing.NoReturn:
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func) -> typing.Any:
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


classproperty = ClassPropertyDescriptor.new


"""Declares :class:`Instance`."""
import inspect
import typing
from inspect import Signature

from .dependency import Dependency
from .provider import _default as provider
from .provider import Provider


class Instance(Dependency):
    """A :class:`cbra.Dependency` implementation that returns a class instance,
    with the dependencies injected into its constructor.

    If the `dependency` argument is a callable, then invoke the callable when
    the :class:`Instance` is resolved at run time. Otherwise, it is assumed
    to be a string that points to a dependency that was injected at boot using
    :meth:`cbra.Application.inject()`.
    """
    __module__: str = 'cbra.ext.ioc'
    _dependency: typing.Union[str, typing.Type[typing.Any]]
    _factory: typing.Optional[typing.Type[typing.Any]]
    _provider: typing.Optional[Provider]

    @property
    def factory(self) -> typing.Callable[..., typing.Any]:
        """Return the factory callable that produces an instance."""
        if self._factory is None:
            if isinstance(self._dependency, str):
                self._dependency = self.provider.resolve(self._dependency)
            self._factory = self._dependency
        return self._factory

    @property
    def provider(self) -> Provider:
        return self.get_provider()

    @property
    def __signature__(self) -> Signature:
        return inspect.signature(self.factory)

    def __init__(self,
        dependency: typing.Union[str, typing.Type[typing.Any]],
        use_cache: bool = True,
        provider: typing.Optional[Provider] = None
    ):
        """Resolve an instance of `dependency` at runtime.

        If `dependency` is a string, then it is assumed to have been priorly
        injected through the :mod:`cbra.ext.ioc` framework. Else, it should be a
        class (instance of :class:`type`).
        """
        self._dependency = dependency
        self._factory = None
        self._provider = provider
        super().__init__(use_cache=use_cache)

    def get_provider(self) -> Provider:
        return self._provider or provider

    async def resolve(self, *args, **kwargs) -> typing.Any:
        """Return an instance of the class provided to the constructor."""
        return self.factory(*args, **kwargs) # pylint: disable=E1102

    async def __call__(self, *args, **kwargs) -> typing.Any:
        return await self.resolve(*args, **kwargs)



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.exceptions import CanonicalException


class SessionRequired(CanonicalException):
    """Raised when a session needs to be established prior to
    invoking an endpoint.
    """
    __module__: str = 'cbra.session'
    http_status_code: int = 401
    code: str = 'SESSION_REQUIRED'
    message: str = (
        "This endpoint requires the client to establish a session prior "
        "to invocation."
    )


"""Declares :class:`IUpstreamReturnHandler`."""
import fastapi
import pydantic

from .authorizationrequestparameters import AuthorizationRequestParameters
from .istorage import IStorage
from .iupstreamprovider import IUpstreamProvider


class IUpstreamReturnHandler:
    __module__: str = 'cbra.ext.oauth2.types'
    provider: str
    summary: str
    query_model: type[pydantic.BaseModel] | None = None
    storage: IStorage

    @classmethod
    def add_to_router(
        cls,
        *,
        app: fastapi.FastAPI | fastapi.APIRouter,
        base_path: str
    ) -> None:
        raise NotImplementedError

    async def get_authorization_request(
        self,
        provider: IUpstreamProvider,
        request: fastapi.Request
    ) -> AuthorizationRequestParameters:
        """Return the authorization request for which we are handling
        a redirect from the external identity provider.
        """
        raise NotImplementedError


# pylint: skip-file
import httpx
import pytest
import pytest_asyncio
from ckms.core import Keychain

from cbra import Application
from cbra.ext.oauth2 import AuthorizationServer


class TestMetadataEndpoint:
    metadata_url: str = '/.well-known/oauth-authorization-server'

    @pytest_asyncio.fixture # type: ignore
    async def app(
        self,
        server: AuthorizationServer,
        keychain: Keychain
    ) -> Application:
        await keychain
        app = Application(
            keychain=keychain
        )
        app.add(server, base_path='/oauth2') # type: ignore
        await app.on_startup()
        await server.setup_authorization_server()
        return app

    @pytest.mark.asyncio
    async def test_get_metadata(self, client: httpx.AsyncClient):
        response = await client.get(self.metadata_url) # type: ignore
        assert response.status_code == 200

    @pytest.mark.asyncio
    async def test_metadata_endpoint_accepts_origin(self, client: httpx.AsyncClient):
        origin = "https://www.example.com"
        response = await client.get( # type: ignore
            url=self.metadata_url,
            headers={'Origin': origin}
        )
        assert response.status_code == 200
        assert response.headers.get('access-control-allow-methods') == "GET"
        assert response.headers.get('access-control-allow-origin') == origin


# pylint: skip-file
import asyncio.coroutines
import inspect
import typing

import fastapi
from ckms.core import Keychain
from ckms.jose import PayloadCodec

from cbra.ext import ioc
from .exceptions import Error
from .types import IClientRepository
from .types import IOpenAuthorizationServer
from .types import IOIDCTokenBuilder
from .types import IStorage
from .types import ISubjectRepository
from .types import IUpstreamProvider


__all__ = [
    'ClientRepository',
    'CurrentServerMetadata',
    'CurrentUpstreamProvider',
    'DownstreamProvider',
    'LocalIssuer',
    'TransientStorage',
    'Server',
    'ServerCodec',
    'ServerJWKS',
    'SubjectRepository',
    'TokenEndpointURL',
    'UpstreamProvider',
]

def _get_server(request: fastapi.Request) -> IOpenAuthorizationServer:
    return request.scope['oauth2']


def get_codec(request: fastapi.Request) -> PayloadCodec:
    keychain: Keychain = request.scope['oauth2'].keychain
    return PayloadCodec(
        signer=keychain.tagged('oauth2-sign'),
        decrypter=keychain.tagged('oauth2-encrypt')
    )


def get_downstream_provider(request: fastapi.Request, provider: str) -> IUpstreamProvider:
    return request.app.providers[provider]


def get_upstream_provider(
    request: fastapi.Request,
    provider: str | None = fastapi.Query(
        default=None,
        title="Provider",
        description=(
            "Specifies the upstream identity provider that establishes "
            "the identity of the resource owner. Additional parameters "
            "may be required to construct the request to the upstream "
            "identity provider."
        )
    )
) -> IUpstreamProvider | None:
    """Return a :class:`~cbra.ext.oauth2.types.IUpstreamProvider` implementation
    that is specified by the request parameter, or ``None`` if the parameter was
    not present.
    """
    if provider is not None and provider not in request.app.providers:
        raise Error(
            error="invalid_request",
            error_description=(
                "The upstream identity provider specified by the `provider` "
                "parameter is not known to the authorization server."
            ),
            mode='client'
        )
    return request.app.providers[provider] if provider else None


def get_local_issuer(request: fastapi.Request) -> str:
    return f'{request.url.scheme}://{request.url.netloc}'


def get_server_jwks(request: fastapi.Request):
    keychain = request.app.keychain.tagged('oauth2')
    return keychain.as_jwks(private=False)


ClientRepository: IClientRepository = ioc.instance(
    'cbra.ext.oauth2.ClientRepository'
)

Server = fastapi.Depends(_get_server)


async def get_server_metadata(
    request: fastapi.Request,
    server: IOpenAuthorizationServer = Server
) -> typing.Any:
    return await server.get_metadata(request) # type: ignore


def get_token_endpoint(request: fastapi.Request) -> str:
    return request.url_for('oauth2.token')


def CurrentUpstreamProvider(name: str) -> IUpstreamProvider:
    def get_upstream_provider(request: fastapi.Request):
        return request.app.providers[name]
    return fastapi.Depends(get_upstream_provider)


CurrentServerMetadata = fastapi.Depends(get_server_metadata)

LocalIssuer = fastapi.Depends(get_local_issuer)

OIDCTokenBuilder: IOIDCTokenBuilder = ioc.instance("OIDCTokenBuilder")

ServerCodec: PayloadCodec = fastapi.Depends(get_codec)

ServerJWKS = fastapi.Depends(get_server_jwks)

SubjectRepository: ISubjectRepository = ioc.instance(
    'cbra.ext.oauth2.SubjectRepository'
)

TokenEndpointURL: str = fastapi.Depends(get_token_endpoint)

TransientStorage: IStorage = ioc.instance(
    'cbra.ext.oauth2.TransientStorage'
)

DownstreamProvider: IUpstreamProvider = fastapi.Depends(get_downstream_provider)

UpstreamProvider: IUpstreamProvider = fastapi.Depends(get_upstream_provider)


class ServerDependant:
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore
    attname: str
    call: typing.Callable[..., typing.Any] | None

    @property
    def __signature__(self) -> inspect.Signature:
        assert self.call is not None # nosec
        return inspect.signature(self.call)

    def __init__(self, attname: str):
        self.attname = attname
        self.call = None

    def add_to_server(self, server: typing.Any) -> None:
        """Adds the dependant to the authorization server."""
        self.call = getattr(server, self.attname)
        signature = inspect.signature(self.call)
        for param in signature.parameters.values():
            if not isinstance(param.default, ServerDependant):
                continue
            param.default.add_to_server(server)


"""Declares :class:`Provider`."""
import typing
from typing import Any

from .models import DependencyList


class Provider:
    """Provides an interface to load and resolve dependencies."""
    __module__: str = 'cbra.ext.ioc'
    DependencySatisfied = type('DependencySatisfied', (Exception,), {})
    _injected: dict[str, Any]

    def __init__(self):
        self._injected = {}

    def is_satisfied(self, name: str) -> bool:
        """Return a boolean indicating if the dependency is satisfied."""
        return name in self._injected # type: ignore

    async def load_many(self, dependencies: DependencyList):
        """Loads all dependencies in the :class:`DependencyList`."""
        for dep, resolved in await dependencies.resolve():
            self.provide(
                name=dep.spec.name,
                value=resolved,
                force=dependencies.force
            )

    def provide(
        self,
        name: str,
        value: object,
        force: bool = False
    ) -> None:
        """Register Python object `value` as a dependency under the key `name` and
        return the object.

        The `force` argument indicates if any existing dependency under `name`
        must be overwrriten. If `force` is ``False``, an exception is raised if
        `name` is already provided.
        """
        if not force and name in self._injected: # pragma: no cover
            raise self.DependencySatisfied(name)
        self._injected[name] = value

    def resolve(self, name: str) -> typing.Any:
        """Resolve a priorly injected dependency by its name."""
        return self._injected[name]


_default: Provider = Provider()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
from aorta.transport import ITransport

from cbra.ext import ioc


class MessagePublisher(aorta.MessagePublisher):

    def __init__(
        self,
        transport: ITransport = ioc.inject('MessageTransport')
    ):
        super().__init__(transport)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestProductCreated(BaseMessageTest):
    event_type: str = 'created'
    resource_type: str = 'product'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "4df0a951a314adc554e0f2016102e020",
                'x-wc-webhook-event': "created",
                'x-wc-webhook-id': "58",
                'x-wc-webhook-resource': "product",
                'x-wc-webhook-signature': "xDHr5XRGdiaAVKotCrxsHHfLXWaO+MzXOmBVKs1uNWQ=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "product.created",
            },
            'content': (
                b'{"id":74,"name":"product.created","slug":"product-create'
                b'd","permalink":"https:\\/\\/test-wordpress.molano.nl\\/prod'
                b'uct\\/product-created\\/","date_created":"2022-06-03T17:22'
                b':45","date_created_gmt":"2022-06-03T17:22:45","date_modi'
                b'fied":"2022-06-03T17:22:45","date_modified_gmt":"2022-06'
                b'-03T17:22:45","type":"simple","status":"publish","featur'
                b'ed":false,"catalog_visibility":"visible","description":"'
                b'","short_description":"","sku":"","price":"","regular_pr'
                b'ice":"","sale_price":"","date_on_sale_from":null,"date_o'
                b'n_sale_from_gmt":null,"date_on_sale_to":null,"date_on_sa'
                b'le_to_gmt":null,"on_sale":false,"purchasable":false,"tot'
                b'al_sales":0,"virtual":false,"downloadable":false,"downlo'
                b'ads":[],"download_limit":-1,"download_expiry":-1,"extern'
                b'al_url":"","button_text":"","tax_status":"taxable","tax_'
                b'class":"","manage_stock":false,"stock_quantity":null,"ba'
                b'ckorders":"no","backorders_allowed":false,"backordered":'
                b'false,"low_stock_amount":null,"sold_individually":false,'
                b'"weight":"","dimensions":{"length":"","width":"","height'
                b'":""},"shipping_required":true,"shipping_taxable":true,"'
                b'shipping_class":"","shipping_class_id":0,"reviews_allowe'
                b'd":true,"average_rating":"0.00","rating_count":0,"upsell'
                b'_ids":[],"cross_sell_ids":[],"parent_id":0,"purchase_not'
                b'e":"","categories":[{"id":15,"name":"Uncategorized","slu'
                b'g":"uncategorized"}],"tags":[],"images":[],"attributes":'
                b'[],"default_attributes":[],"variations":[],"grouped_prod'
                b'ucts":[],"menu_order":0,"price_html":"","related_ids":[3'
                b'9,46,14,44,13],"meta_data":[],"stock_status":"instock","'
                b'_links":{"self":[{"href":"https:\\/\\/test-wordpress.molan'
                b'o.nl\\/wp-json\\/wc\\/v3\\/products\\/74"}],"collection":[{"h'
                b'ref":"https:\\/\\/test-wordpress.molano.nl\\/wp-json\\/wc\\/v'
                b'3\\/products"}]}}'
            )
        }


"""Declares :class:`ErrorProducer.`"""
import typing

from .exceptions import Error


class ErrorProducer:
    __module__: str = 'ckms.ext.oauth2'

    def reject_request(self, description: typing.Optional[str] = None):
        """Reject the token exchange request with an error state of
        `invalid_request`.
        """
        self._reject("invalid_request", description)

    def reject_target(self, description: typing.Optional[str] = None):
        """Reject the token exchange request with an error state of
        `invalid_target`.
        """
        self._reject("invalid_target", description)

    def _reject(self, code, description, url=None):
        raise Error(
            error=code,
            error_description=description,
            error_uri=url
        )


"""Declares :class:`TokenIssuer`."""
import functools
import inspect
import typing

from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from ckms.types import Malformed

from cbra.conf import settings
from .exceptions import InvalidClient
from .exceptions import InvalidGrant
from .exceptions import InvalidScope
from .exceptions import ForbiddenScope
from .exceptions import AssertionReplayed
from .params import CurrentServerMetadata
from .params import LocalIssuer
from .params import OIDCTokenBuilder
from .params import ServerCodec
from .params import SubjectRepository
from .params import TransientStorage
from .types import AuthorizationCodeGrant
from .types import BaseGrant
from .types import ClientCredentialsGrant
from .types import GrantType
from .types import IClient
from .types import IOIDCTokenBuilder
from .types import IStorage
from .types import ISubject
from .types import ISubjectRepository
from .types import ITokenIssuer
from .types import JWTBearerAssertionGrant
from .types import ServerMetadata
from .types import TokenResponse
from .utils import classproperty # type: ignore


SIGNING_KEY: str = getattr(settings, 'OAUTH_SIGNING_KEY', 'sig')


class TokenIssuer(ITokenIssuer):
    """Issues a token for various grant types."""
    __module__: str = 'cbra.ext.oauth2'

    #: The transient storage used.
    storage: IStorage

    #: OpenID Connect token builder
    openid: IOIDCTokenBuilder

    @classproperty
    def grant_types_supported(cls) -> typing.List[str]:
        return [
            GrantType.authorization_code.value,
            GrantType.client_credentials.value,
            GrantType.jwt_bearer.value
        ]

    def __init__(
        self,
        subjects: ISubjectRepository = SubjectRepository,
        codec: PayloadCodec = ServerCodec,
        metadata: ServerMetadata = CurrentServerMetadata,
        issuer: str = LocalIssuer,
        storage: IStorage = TransientStorage,
        openid: IOIDCTokenBuilder = OIDCTokenBuilder
    ):
        self.subjects = subjects
        self.codec = codec
        self.metadata = metadata
        self.issuer = issuer
        self.openid = openid
        self.storage = storage

    def is_local_issued(self, claims: JSONWebToken) -> bool:
        """Return a boolean indicating if the claimset was issued
        by the local server.
        """
        return self.issuer == claims.iss

    @functools.singledispatchmethod
    async def grant(
        self,
        dto: BaseGrant
    ) -> TokenResponse:
        """Dispatches a request to the **Token Endpoint** to the
        appropriate handler for its ``grant_type``.
        """
        raise NotImplementedError(dto.grant_type)

    @grant.register
    async def grant_authorization_code(
        self,
        grant: AuthorizationCodeGrant
    ) -> TokenResponse:
        client = grant.get_client()
        _, request = await self.storage.get_code(grant.code)

        assert request.is_authenticated() # nosec
        subject = await request.get_subject(client, self.subjects)
        request.validate_grant(client, subject, grant)
        self.validate_grant(
            client=client,
            subject=subject,
            grant=grant
        )


        assert subject is not None # nosec
        at = await client.issue_token(
            codec=self.codec,
            using=SIGNING_KEY,
            issuer=self.metadata.issuer,
            audience=grant.get_audience(),
            subject=subject,
            ttl=self.default_ttl,
            scope=request.scope
        )
        return TokenResponse(
            access_token=at,
            token_type="Bearer",
            expires_in=self.default_ttl,
            id_token=await self.openid.build(
                signing_key=SIGNING_KEY,
                client=client,
                subject=subject,
                grant=grant,
                scope=request.scope,
                request=request,
                access_token=at
            ),
            state=request.state,
            refresh_token=None
        )

    @grant.register
    async def grant_client_credentials(
        self,
        grant: ClientCredentialsGrant
    ) -> TokenResponse:
        # Client credentials grant is only available for confidential
        # clients.
        client = grant.get_client()
        if not client.is_confidential():
            raise InvalidClient(
                error="invalid_client",
                error_description=(
                    "The \"client_credentials\" grant is only available "
                    "for confidential clients."
                )
            )
        if grant.scope is not None and not client.allows_scope(grant.scope):
            raise ForbiddenScope

        # If the request does not include a "resource" parameter, the authorization
        # server MUST use a default resource indicator in the "aud" claim (RFC 9068,
        # Section 3).
        if not grant.resource:
            grant.resource.add(self.issuer) # type: ignore
        if len(grant.resource) > 1 and not client.can_issue_multiple():
            raise InvalidGrant(
                error_description=(
                    "The client refuses to issue access tokens for multiple "
                    "audiences."
                )
            )
        if not client.allows_audience(self.issuer, grant.resource):
            raise InvalidGrant(
                error_description=(
                    "The resource(s) specified in the request is not allowed "
                    "by the client."
                )
            )
        return TokenResponse(
            access_token=await client.issue_token(
                codec=self.codec,
                algorithm='EdDSA',
                using='sig',
                issuer=self.issuer,
                audience=grant.get_audience(),
                subject=client.as_subject(),
                ttl=self.default_ttl,
                scope=set(grant.scope or [])
            ),
            token_type="Bearer",
            expires_in=self.default_ttl,
            id_token=None,
            state=None,
            refresh_token=None
        )

    @grant.register
    async def grant_jwt_bearer(
        self,
        dto: JWTBearerAssertionGrant
    ) -> TokenResponse:
        client = dto.get_client()
        try:
            jws, claims = await self.codec.jwt(dto.assertion)
        except (Malformed, ValueError, TypeError):
            raise InvalidGrant(
                error_description=(
                    "The assertion is malformed and could not be interpreted "
                    "as a JSON Web Token (JWT)."
                )
            )
        claims.verify(
            audience={self.metadata.token_endpoint},
            max_age=self.max_assertion_age,
            required={'iss', 'aud', 'sub', 'iat', 'exp', 'nbf', 'jti'}
        )

        # Check if the assertion is not being replayed.
        if await self.storage.consume(claims):
            raise AssertionReplayed

        # There are two cases that need to be handled here, either the
        # claims are self-signed (a public key is preregistered for the
        # subject specified in the `sub` claim), or the claims were
        # signed by a trusted third-party.
        subject = await self.subjects.get(
            client_id=client.client_id,
            subject_id=typing.cast(typing.Union[int, str], claims.sub)
        )
        if dto.scope is not None and not client.allows_scope(dto.scope):
            raise ForbiddenScope
        if not subject.allows_scope(set(dto.scope or [])):
            raise InvalidScope

        # Lookup the JWKS of the issuer as indicated by the `iss` claim,
        # or use the keys pre-registered by the resource owner if the claims
        # are self-signed.
        selfsigned = claims.is_selfsigned()
        if not selfsigned and not self.is_local_issued(claims):
            # The assertion was issued and signed by a third-party.
            if not self.allow_jwks_lookups:
                raise InvalidGrant(
                    error_description=(
                        "The assertion was signed by a third-party, but it's "
                        "public keys are not known to the server and it does "
                        "not allow import using the JWKS URI."
                    )
                )
            if claims.iss not in self.trusted_issuers:
                raise InvalidGrant(
                    error_description=(
                        "The assertion was issued and signed by an issuer that is "
                        "not trusted by the server."
                    )
                )
            raise NotImplementedError
        elif not selfsigned:
            # The assertion was issued and signed by this server.
            raise InvalidGrant(
                error_description=(
                    "The assertion was issued and signed by an issuer that is "
                    "not trusted by the server."
                )
            )
        else:
            verifier = subject

        verified = verifier.verify(jws)
        if inspect.isawaitable(verified):
            verified = await verified
        if not verified:
            raise InvalidGrant(
                error_description=(
                    "The signature of the JWT assertion did not validate. Make "
                    "sure that the signing keys are pre-registered by the "
                    "resource owner."
                )
            )

        # If the request does not include a "resource" parameter, the authorization
        # server MUST use a default resource indicator in the "aud" claim (RFC 9068,
        # Section 3).
        if not dto.resource:
            dto.resource.add(self.issuer) # type: ignore
        if not client.allows_audience(self.issuer, set(dto.resource or [])):
            raise InvalidGrant(
                error_description=(
                    "Any or all of the audiences are not allowed by the client."
                )
            )

        token = await client.issue_token(
            codec=self.codec,
            algorithm='EdDSA',
            using='sig',
            issuer=self.issuer,
            audience=dto.get_audience(),
            subject=subject,
            ttl=self.default_ttl,
            scope=set(dto.scope or [])
        )
        return TokenResponse(
            access_token=token,
            token_type="Bearer",
            expires_in=self.default_ttl,
            id_token=None,
            state=None,
            refresh_token=None
        )

    def validate_grant(
        self,
        client: IClient,
        subject: ISubject | None,
        grant: BaseGrant
    ) -> None:
        """Validates the parameters provided by the grant and sets defaults."""
        # If the request does not include a "resource" parameter, the authorization
        # server MUST use a default resource indicator in the "aud" claim (RFC 9068,
        # Section 3).
        if not grant.resource:
            grant.resource.add(self.issuer) # type: ignore
        if not client.allows_audience(self.issuer, set(grant.resource or [])):
            raise InvalidGrant(
                error_description=(
                    "Any or all of the audiences are not allowed by the client."
                )
            )
        grant.validate_grant(client, subject)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`EndpointResolver`."""
from collections import OrderedDict
from inspect import getmembers
from inspect import Parameter
from inspect import Signature
from typing import get_type_hints
from typing import Any

import fastapi.params


class EndpointResolver:
    """Inspect the attributes of a class that are :mod:`fastapi.params.Depends`
    instances and ensure that they are resolved prior to invoking the request
    handler.
    """
    endpoint: Any
    params: OrderedDict[str, Parameter]

    @property
    def __signature__(self) -> Signature:
        return Signature(
            parameters=list(self.params.values()),
            return_annotation=None
        )

    def __init__(self, endpoint: Any):
        self.endpoint = endpoint
        self.params = OrderedDict()
        self.hints = get_type_hints(self.endpoint)
        for attname, value in getmembers(self.endpoint):
            if not isinstance(value, fastapi.params.Depends):
                continue
            self.params[attname] = Parameter(
                kind=Parameter.POSITIONAL_ONLY,
                name=attname,
                default=value,
                annotation=self.hints.get(attname)
            )

    def __call__(self, **kwargs: Any) -> None:
        for attname, value in kwargs.items():
            assert attname in self.params # nosec
            setattr(self.endpoint, attname, value)


"""Declares :class:`BodyConsumer`."""
import typing

from fastapi import Request

from .types import IContentNegotiation
from .negotiation import DefaultContentNegotiation
from .parsers import IParser
from .parsers import FormParser
from .parsers import JSONParser


class BodyConsumer:
    """Mixin class that provides an interface to consume the body of an
    HTTP request.
    """
    #: The content negiotiation procedure. Subclasses may override
    #: this attribute to provide their own implementation.
    #: See also :class:`~unimatrix.ext.webapi.negotiation.IContentNegotiation`.
    negotiator: IContentNegotiation = DefaultContentNegotiation(request=None)

    #: The list of request body parsers in order of preference.
    parsers: typing.List[IParser] = []

    #: Default parsers for the specific :class:`BodyConsumer`
    #: implementations.
    default_parsers: typing.List[IParser] = [
        JSONParser,
        FormParser
    ]

    async def get_body(self, request: Request = None) -> typing.Any:
        """Parse and return the request body."""
        # Do not parse the body for methods that do not commonly use them.
        #
        #   HTTP request bodies are theoretically allowed for all methods
        #   except TRACE, however they are not commonly used except in PUT,
        #   POST and PATCH. Because of this, they may not be supported properly
        #   by some client frameworks, and you should not allow request bodies
        #   for GET, DELETE, TRACE, OPTIONS and HEAD methods.
        #
        # (from OpenStack API Special Interest Group documentation)
        request = request or self.request
        if request.method in {"GET", "HEAD", "TRACE", "OPTIONS", "DELETE"}:
            return None
        media_type = request.headers.get('Content-Type')
        content_length = request.headers.get('Content-Length') or ''
        if not str.isdigit(content_length) or int(content_length) == 0: # pragma: no cover
            return None
        parser = self.negotiator.select_parser(request, self.get_parsers())
        return await parser.parse(request, media_type)

    def get_parsers(self) -> typing.List[IParser]:
        """Return the list of parsers used to deserialize the request body."""
        return self.parsers + self.default_parsers



"""Declares :class:`JSONQueryParameter`."""
import urllib.parse
from typing import Any
from typing import Generator


class JSONQueryParameter(str):

    @classmethod
    def __get_validators__(cls) -> Generator[Any, None, None]:
        yield cls.validate

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None: # pragma: no cover
        field_schema.update({
            "type": "str"
        })

    @classmethod
    def validate(
        cls,
        value: str | None
    ) -> "JSONQueryParameter":
        """Unquote the input data and deserialize as JSON."""
        if isinstance(value, cls) or value is None:
            return cls(value or '{}')
        return cls(urllib.parse.unquote(value))
        


# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`SubjectPrincipal`."""
import dataclasses

from cbra.types import IPrincipal


@dataclasses.dataclass
class SubjectPrincipal(IPrincipal):
    """A :class:`~cbra.types.IPrincipal` implementation that simply
    references a subject by its identifier.
    """
    sub: int | str

    def is_authenticated(self) -> bool:
        return True